from dataclasses import asdict
from dataclasses import is_dataclass
from logging import getLogger
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Mapping
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import get_type_hints

from botox import Injector

from watermill.message_brokers.message_broker import EndOfStream
from watermill.message_brokers.message_broker import EndOfStreamError
from watermill.message_brokers.message_broker import MessageBroker
from watermill.message_brokers.message_broker import RightJoinedStreamRanAhead
from watermill.message_brokers.message_broker import StreamTimeoutError
from watermill.message_brokers.message_broker import StreamsShift
from watermill.stream_join import JoinTree
from watermill.stream_join import JoinTreeNode
from watermill.stream_join import compare_streams
from watermill.stream_join import get_injected_kwargs
from watermill.stream_join import get_join_data_mapping
from watermill.stream_join import get_tree_pairs
from watermill.message_brokers.message_broker import DataKey
from watermill.stream_types import JoinedStreamType
from watermill.stream_types import ResultStreamType
from watermill.stream_types import StreamType
from watermill.windows import Window

logger = getLogger(__name__)


def _default_serializer(dataclass_value) -> dict:
    assert is_dataclass(dataclass_value)
    return asdict(dataclass_value)


def _default_deserializer(cls):
    def wrapped_de(value: dict):
        return cls(**value)
    return wrapped_de


class WaterMill:
    def __init__(
            self,
            broker: MessageBroker,
            process_func: Callable[[StreamType, List[Mapping[StreamType, JoinedStreamType]]], Union[ResultStreamType, None]],
            stream_cls: Union[Type[StreamType], Window] = None,
            join_tree: JoinTree = None,
            serializers: Mapping[Type, Callable] = None,
            deserializers: Mapping[Type, Callable] = None,
            injector: Injector = None,
            output_broker: Optional[MessageBroker] = None,
            explicit_return_types: List[StreamType] = None
    ):
        assert stream_cls or join_tree

        self._injector = injector or Injector()
        self._injector.prepare(WaterMill, self)
        self._broker = broker
        self._output_broker = output_broker or broker
        self._window_func = None
        self._eos = False
        self._data_keys_to_send = set()
        self._explicit_return_types = explicit_return_types
        type_hints = get_type_hints(process_func)
        self._return_type = type_hints.get('return')
        if self._return_type == NoReturn or self._return_type == type(None):
            self._return_type = None
        elif str(self._return_type).startswith('typing.Union'):
            for return_type_item in self._return_type.__reduce__()[1][1]:
                if return_type_item != NoReturn:
                    self._return_type = return_type_item
                    break

        if stream_cls:
            if isinstance(stream_cls, Window):
                self._join_tree = JoinTree(
                    root_node=JoinTreeNode(
                        user_type=stream_cls.cls,
                    )
                )
                self._window_func = stream_cls.window_func
            else:
                self._join_tree = JoinTree(
                    root_node=JoinTreeNode(
                        user_type=stream_cls,
                    )
                )
        else:
            self._join_tree = join_tree

        self._process_func = process_func
        self._serializers: Mapping[Type, Callable] = serializers or {}
        self._deserializers: Mapping[Type, Callable] = deserializers or {}

    def run(self):
        if self._join_tree is None:
            return

        join_tree = self._join_tree
        serializer = self._serializers.get(self._return_type, _default_serializer) if self._return_type else None
        deserializer = self._deserializers.get(join_tree.root_node.user_type, _default_deserializer(join_tree.root_node.user_type))
        processing_func_type_hints = get_type_hints(self._process_func)
        current_key = None
        self._injector.prepare(DataKey, lambda: current_key)

        window_key = None
        root_elements = []
        join_data_mappings = []
        process_results = []
        window_func = self._window_func
        if join_tree.root_node.window_func:
            window_func = join_tree.root_node.window_func

        window_comparator = lambda next_element: StreamsShift.Equal if (window_key is None or window_key == window_func(next_element)) else StreamsShift.Less

        while True:
            root_element_data = None
            eos = False
            try:
                root_element_data = next(self._broker.get_elements(join_tree.root_node.user_type, comparator=window_comparator))
                current_key = self._broker.get_data_key()
            except RightJoinedStreamRanAhead:
                injected_kwargs = get_injected_kwargs(processing_func_type_hints, self._injector)
                if current_key:
                    self._data_keys_to_send.add(current_key.key)
                process_result = self._process_func(root_elements, *join_data_mappings, **injected_kwargs)
                process_results, eos = _handle_result(process_result, process_results, eos)
                root_elements.clear()
                join_data_mappings.clear()
                window_key = None
            except StreamTimeoutError:
                new_window_key = window_func(None)

                assert new_window_key is not None
                if not window_key or window_key == new_window_key:
                    window_key = new_window_key
                    continue
                else:
                    if not root_elements:
                        continue

                    injected_kwargs = get_injected_kwargs(processing_func_type_hints, self._injector)
                    if current_key:
                        self._data_keys_to_send.add(current_key.key)
                    process_result = self._process_func(root_elements, *join_data_mappings, **injected_kwargs)
                    process_results, eos = _handle_result(process_result, process_results, eos)
                    root_elements.clear()
                    join_data_mappings.clear()
                    window_key = new_window_key
            except EndOfStreamError:
                if self._return_type:
                    self._output_broker.send_eos(self._return_type)
                self._eos = True
                return

            join_data_mapping = []
            if root_element_data:
                root_user_element = deserializer(root_element_data)
                if join_tree.pairs:
                    parent_children_map: Dict[StreamType, List[JoinedStreamType]] = {}
                    try:
                        self._get_joined_elements(join_tree.root_node, join_tree, root_element_data, root_user_element, parent_children_map)
                        if current_key:
                            current_key =  self._broker.get_data_key()
                    except RightJoinedStreamRanAhead as exc:
                        if not window_func:
                            logger.warning(f'Joined right stream {exc.stream_type} ran ahead. Skipping root element')
                            continue
                    except EndOfStreamError:
                        self._eos = True
                        if parent_children_map:
                            eos = True
                        elif self._return_type:
                            self._output_broker.send_eos(self._return_type)
                            self._broker.commit()
                            return

                    join_data_mapping = get_join_data_mapping(processing_func_type_hints, parent_children_map)

                if window_func:
                    new_window_key = window_func(root_element_data)
                    assert new_window_key is not None
                    root_elements.append(root_user_element)
                    if not join_data_mappings:
                        join_data_mappings = join_data_mapping
                    else:
                        assert len(join_data_mapping) == len(join_data_mappings)
                        temp_join_data_mappings = []
                        for old, new in zip(join_data_mappings, join_data_mapping):
                            old.update(new)
                            temp_join_data_mappings.append(old)
                        join_data_mappings = temp_join_data_mappings

                    if eos:
                        injected_kwargs = get_injected_kwargs(processing_func_type_hints, self._injector)
                        if current_key:
                            self._data_keys_to_send.add(current_key.key)
                        process_result = self._process_func(root_elements, *join_data_mappings, **injected_kwargs)
                        process_results, eos = _handle_result(process_result, process_results, eos)
                    else:
                        window_key = new_window_key
                        continue
                else:
                    injected_kwargs = get_injected_kwargs(processing_func_type_hints, self._injector)
                    if current_key:
                        self._data_keys_to_send.add(current_key.key)
                    process_result = self._process_func(root_user_element, *join_data_mapping, **injected_kwargs)
                    process_results, eos = _handle_result(process_result, process_results, eos)

            assert process_results or not self._return_type

            if self._return_type:
                for result in process_results:
                    self._output_broker.send(self._return_type, serializer(result))
                process_results.clear()
                self._broker.commit()
            if eos:
                self._eos = True
                if self._return_type:
                    self._output_broker.send_eos(self._return_type)
                    self._broker.commit()
                return

    def save_results(self, data_key: DataKey, data: Union[StreamType, List[StreamType], None, EndOfStream]):
        if data:
            if isinstance(data, list):
                for data_item in data:
                    return_type = type(data_item)
                    local_serializer = self._serializers.get(return_type, _default_serializer)
                    self._output_broker.send(return_type, local_serializer(data_item))
            elif isinstance(data, EndOfStream):
                self._eos = True
                self._data_keys_to_send = {data_key.key}
            else:
                return_type = type(data)
                local_serializer = self._serializers.get(return_type, _default_serializer)
                self._output_broker.send(return_type, local_serializer(data))

        self._broker.commit(data_key)
        try:
            self._data_keys_to_send.remove(data_key.key)
        except KeyError:
            logger.exception(f'Failed to remove data key {data_key}')

        if not self._data_keys_to_send and self._eos:
            for explicit_return_type in self._explicit_return_types:
                self._output_broker.send_eos(explicit_return_type)

    def _get_joined_elements(
            self,
            join_node: JoinTreeNode,
            join_tree: JoinTree,
            left_element: dict,
            left_deserialized_element: StreamType,
            parent_children_map: Dict[StreamType, List[JoinedStreamType]],
    ):
        eos = False
        eos_exc = None
        for join_pair in get_tree_pairs(join_tree, join_node):
            right_join_node = join_pair.right_node
            left_expression = join_pair.left_expression
            right_expression = join_pair.right_expression

            try:
                for element_data in self._broker.get_elements(
                        right_join_node.user_type,
                        comparator=lambda right_element: compare_streams(left_expression(left_element), right_expression(right_element))
                ):
                    assert element_data is not None
                    deserialized_element = self._deserializers.get(right_join_node.user_type, _default_deserializer(right_join_node.user_type))(element_data)
                    parent_children_map.setdefault(left_deserialized_element, []).append(deserialized_element)

                    self._get_joined_elements(right_join_node, join_tree, element_data, deserialized_element, parent_children_map)
            except EndOfStreamError as exc:
                eos = True
                eos_exc = exc
                continue

        if eos:
            raise eos_exc


def _handle_result(process_result: Union[StreamType, StopIteration, None], process_results: list, eos: bool) -> Tuple[list, bool]:
    if not process_result or process_result is None:
        return process_results, eos

    if isinstance(process_result, StopIteration):
        return process_results, True

    if isinstance(process_result, Generator):
        for element in process_result:
            assert element
            if isinstance(element, StopIteration):
                return process_results, True
            process_results.append(element)
    else:
        process_results.append(process_result)
    return process_results, eos



import enum
from dataclasses import field
from uuid import UUID
from uuid import uuid4
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type
from typing import TypeVar


StreamType = TypeVar('StreamType')
JoinedStreamType = TypeVar('JoinedStreamType')


class EndOfStreamError(Exception):
    def __init__(
            self,
            stream_type: Type[StreamType],
    ):
        super().__init__()
        self._stream_type = stream_type


class StreamTimeoutError(Exception):
    def __init__(
            self,
            stream_type: Type[StreamType],
    ):
        super().__init__()
        self._stream_type = stream_type


class RightJoinedStreamRanAhead(Exception):
    def __init__(self, stream_type: Type[JoinedStreamType]):
        super().__init__()
        self.stream_type = stream_type


@dataclass
class EndOfStream:
    eos__: bool = True


class StreamsShift(enum.Enum):
    Less = 'Less'
    Equal = 'Equal'
    Greater = 'Greater'


@dataclass
class DataKey:
    key: UUID = field(default_factory=uuid4)


class MessageBroker:
    def send(self, item_type: Type[StreamType], item: dict):
        raise NotImplementedError()

    def send_eos(self, item_type: Type[StreamType]):
        raise NotImplementedError()

    def get_elements(
            self,
            element_type: Type[StreamType],
            comparator: Callable[[Any, Any], StreamsShift] = None
    ) -> dict:
        raise NotImplementedError()

    def commit(self, data_key: Optional[DataKey] = None):
        pass

    def get_data_key(self) -> Optional[DataKey]:
        return None



import json
from logging import getLogger
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Mapping
from typing import Type

from watermill.message_brokers.message_broker import EndOfStreamError
from watermill.message_brokers.message_broker import MessageBroker
from watermill.message_brokers.message_broker import RightJoinedStreamRanAhead
from watermill.message_brokers.message_broker import StreamsShift
from watermill.stream_types import StreamType


logger = getLogger(__name__)


class JsonFileMessageBroker(MessageBroker):
    def __init__(self, stream_paths: Mapping[Type[StreamType], Path]):
        self._stream_paths = stream_paths
        self._stream_file_data: Dict[str, List[dict]] = {}
        self._stream_file_pos: Dict[str, int] = {}
        self._stream_out_data = {}

    def send(self, element_type: Type[StreamType], item: dict):
        self._stream_out_data.setdefault(element_type, []).append(item)

    def send_eos(self, element_type: Type[StreamType]):
        path = self._stream_paths[element_type]
        with path.open('w') as out_file:
            json.dump(self._stream_out_data[element_type], out_file)

    def get_elements(
            self,
            element_type: Type[StreamType],
            comparator: Callable[[Any, Any], StreamsShift] = None
    ) -> dict:
        if not comparator:
            yield self._get_next_element(element_type)
            return

        element = self._get_next_element(element_type)
        while comparator(element) == StreamsShift.Greater:
            logger.warning(f'Skipped stream {element_type} element')
            element = self._get_next_element(element_type)

        if comparator(element) == StreamsShift.Less:
            self._step_back(element_type)
            raise RightJoinedStreamRanAhead(element_type)

        while comparator(element) == StreamsShift.Equal:
            yield element
            element = self._get_next_element(element_type)
        self._step_back(element_type)

    def _get_stream_qualified_name(self, stream_type: Type[StreamType]) -> str:
        return stream_type.__name__

    def _step_back(
            self,
            stream_type: Type[StreamType],
    ):
        stream_name = self._get_stream_qualified_name(stream_type)
        position = self._stream_file_pos[stream_name]
        if position < 1:
            logger.warning(f'Skipping stream {stream_type} element')
            return

        self._stream_file_pos[stream_name] -= 1

    def _get_next_element(
            self,
            stream_type: Type[StreamType],
    ) -> dict:
        stream_name = self._get_stream_qualified_name(stream_type)
        if stream_name not in self._stream_file_data:
            with self._stream_paths[stream_type].open() as f:
                self._stream_file_data[stream_name] = json.load(f)
                self._stream_file_pos[stream_name] = 0

        position = self._stream_file_pos[stream_name]
        if position >= len(self._stream_file_data[stream_name]):
            raise EndOfStreamError(stream_type)
        element = self._stream_file_data[stream_name][position]
        self._stream_file_pos[stream_name] += 1
        return element



import json
import threading
from dataclasses import asdict
from dataclasses import dataclass
from pathlib import Path
from random import uniform
from time import sleep
from time import time
from typing import List

from kafka import KafkaAdminClient
from kafka import KafkaConsumer
from kafka import KafkaProducer
from kafka.admin import NewTopic

from watermill.expressions import function_call
from watermill.message_brokers.kafka_message_broker import KafkaMessageBroker
from watermill.message_brokers.message_broker import EndOfStream
from watermill.mill import WaterMill
from watermill.windows import window


@dataclass(frozen=True, eq=True)
class RobotLocation:
    time: int
    x: int
    y: int


@dataclass(frozen=True, eq=True)
class OccupiedCellsNumber:
    time: int
    cells_number: int


def calculate_cell_occupation(locations: List[RobotLocation]) -> OccupiedCellsNumber:
    time = locations[0].time
    cell_set = set()
    for location in locations:
        cell_set.add((int(location.x) // 50, int(location.y) // 50))
    return OccupiedCellsNumber(
        time=time,
        cells_number=len(cell_set)
    )


def time_based_window_fetch():
    message_broker = KafkaMessageBroker(
        topic_names={
            RobotLocation: 'robot-locations',
            OccupiedCellsNumber: 'cells-occupation',
        },
        poll_timeout=100,
        kafka_url='kafka:9092'
    )

    mill = WaterMill(
        broker=message_broker,
        process_func=calculate_cell_occupation,
        stream_cls=window(
            cls=RobotLocation,
            window_expression=function_call(lambda _: int(time()) // 3)
        )
    )

    produce_data_stream()

    result_consumer = KafkaConsumer(
        bootstrap_servers='kafka:9092',
        group_id='result-consumer',
        auto_offset_reset='earliest',
        enable_auto_commit=True,
    )
    result_consumer.subscribe(['cells-occupation'])
    def get_results():
        for result in result_consumer:
            result_item = json.loads(result.value.decode('utf8'))
            print('result', json.dumps(result_item))
            if 'eos__' in result_item:
                break

    threading.Thread(target=get_results).start()

    mill.run()


def produce_data_stream():
    recreate_topics(['robot-locations', 'cells-occupation'])
    print('Sending data to input stream')

    kafka_producer = KafkaProducer(bootstrap_servers='kafka:9092')

    goods_file_path = Path(__file__).resolve().parent / 'samples' / 'robot-locations.json'
    with goods_file_path.open('r') as f:
        goods = json.load(f)

    def thread_data():
        for goods_item in goods:
            print(goods_item)
            kafka_producer.send('robot-locations', value=json.dumps(goods_item).encode('utf8'))
            sleep(uniform(0, 1.))

        kafka_producer.send('robot-locations', value=json.dumps(asdict(EndOfStream())).encode('utf8'))
        kafka_producer.flush()

    threading.Thread(target=thread_data).start()


def recreate_topics(topics: List[str]):
    kafka_admin_client = KafkaAdminClient(
        bootstrap_servers='kafka:9092',
    )
    try:
        kafka_admin_client.delete_topics(topics)
    except Exception as exc:
        print(exc)
    sleep(0.1)
    try:
        kafka_admin_client.create_topics([NewTopic(name=topic, num_partitions=1, replication_factor=1) for topic in topics])
    except Exception as exc:
        print(exc)


if __name__ == "__main__":
    time_based_window_fetch()



import json
import logging
from dataclasses import asdict
from dataclasses import dataclass
from pathlib import Path
from time import sleep
from typing import List

from kafka import KafkaAdminClient
from kafka import KafkaConsumer
from kafka import KafkaProducer
from kafka.admin import NewTopic

from watermill.message_brokers.kafka_message_broker import KafkaMessageBroker
from watermill.message_brokers.message_broker import EndOfStream
from watermill.mill import WaterMill


@dataclass(frozen=True, eq=True)
class GoodsItem:
    title: str
    quantity: int
    price: int


@dataclass(frozen=True, eq=True)
class UpdatedGoodsItem:
    title: str
    quantity: int
    price: int


def update_goods_price(goods_item: GoodsItem) -> UpdatedGoodsItem:
    return UpdatedGoodsItem(
        title=goods_item.title,
        quantity=goods_item.quantity,
        price=int(goods_item.price * 1.1)
    )


def kafka_stream_example_with_result():
    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger('kafka')
    logger.setLevel(logging.WARNING)

    message_broker = KafkaMessageBroker(
        topic_names={
            GoodsItem: 'goods',
            UpdatedGoodsItem: 'updated_goods',
        },
        kafka_url='kafka:9092'
    )

    mill = WaterMill(
        broker=message_broker,
        process_func=update_goods_price,
        stream_cls=GoodsItem,
    )

    produce_goods_stream()

    mill.run()

    print('Fetching data from resulting stream')
    result_consumer = KafkaConsumer(
        bootstrap_servers='kafka:9092',
        group_id='result-consumer',
        auto_offset_reset='earliest',
        enable_auto_commit=True,
    )
    result_consumer.subscribe(['updated_goods'])
    for result in result_consumer:
        result_item = json.loads(result.value.decode('utf8'))
        print(json.dumps(result_item))
        if 'eos__' in result_item:
            break


def produce_goods_stream():
    recreate_topics(['goods', 'updated_goods'])
    print('Sending data to input stream')

    kafka_producer = KafkaProducer(bootstrap_servers='kafka:9092')

    goods_file_path = Path(__file__).resolve().parent / 'samples' / 'goods.json'
    with goods_file_path.open('r') as f:
        goods = json.load(f)

    for goods_item in goods:
        print(goods_item)
        kafka_producer.send('goods', value=json.dumps(goods_item).encode('utf8'))

    kafka_producer.send('goods', value=json.dumps(asdict(EndOfStream())).encode('utf8'))
    kafka_producer.flush()


def recreate_topics(topics: List[str]):
    kafka_admin_client = KafkaAdminClient(
        bootstrap_servers='kafka:9092',
    )
    try:
        kafka_admin_client.delete_topics(topics)
    except Exception as exc:
        print(exc)
    sleep(0.1)
    try:
        kafka_admin_client.create_topics([NewTopic(name=topic, num_partitions=1, replication_factor=1) for topic in topics])
    except Exception as exc:
        print(exc)


if __name__ == "__main__":
    kafka_stream_example_with_result()



import json
from dataclasses import asdict
from dataclasses import dataclass
from pathlib import Path
from time import sleep
from typing import List
from typing import Mapping

from kafka import KafkaAdminClient
from kafka import KafkaConsumer
from kafka import KafkaProducer
from kafka.admin import NewTopic

from watermill.expressions import get_field
from watermill.message_brokers.kafka_message_broker import KafkaMessageBroker
from watermill.message_brokers.message_broker import EndOfStream
from watermill.mill import WaterMill
from watermill.stream_join import JoinWith
from watermill.stream_join import join_streams
from watermill.stream_types import type_alias
from watermill.windows import window


@dataclass(frozen=True, eq=True)
class RobotLocation:
    robot_id: str
    time: int
    x: int
    y: int


@dataclass(frozen=True, eq=True)
class OccupiedCellsNumber:
    time: int
    cells_number: int


def calculate_cell_occupation(locations: List[RobotLocation], joined_locations: Mapping[RobotLocation, List[RobotLocation]]) -> OccupiedCellsNumber:
    time = locations[0].time
    cell_set = set()
    for location in locations:
        cell_set.add((int(location.x) // 50, int(location.y) // 50))
    return OccupiedCellsNumber(
        time=time,
        cells_number=len(cell_set)
    )


def window_parallel_streams():
    # assume we have N robots
    N = 10

    topic_names = {
        OccupiedCellsNumber: 'cells-occupation',
    }

    for i in range(N):
        topic_names[type_alias(RobotLocation, f"robot-{i}")] = f'robot-locations-{i}'

    message_broker = KafkaMessageBroker(
        topic_names=topic_names,
        kafka_url='kafka:9092'
    )

    mill = WaterMill(
        broker=message_broker,
        process_func=calculate_cell_occupation,
        join_tree=join_streams(
            window(
                cls=type_alias(RobotLocation, f"robot-0"),
                window_expression=get_field('time') // 30
            ),
            *[JoinWith(
                with_type=type_alias(RobotLocation, f"robot-{i}"),
                left_expression=get_field('time') // 30,
                right_expression=get_field('time') // 30
            ) for i in range(1, N)],
        ),
    )

    produce_robot_streams()

    mill.run()

    print('Fetching data from resulting stream')
    result_consumer = KafkaConsumer(
        bootstrap_servers='kafka:9092',
        group_id='result-consumer',
        auto_offset_reset='earliest',
        enable_auto_commit=True,
    )
    result_consumer.subscribe(['cells-occupation'])
    for result in result_consumer:
        result_item = json.loads(result.value.decode('utf8'))
        print(json.dumps(result_item))
        if 'eos__' in result_item:
            break


def produce_robot_streams():
    N = 10
    recreate_topics([*[f'robot-locations-{i}' for i in range(N)], 'cells-occupation'])
    print('Sending data to input stream')

    kafka_producer = KafkaProducer(bootstrap_servers='kafka:9092')

    robots_file_path = Path(__file__).resolve().parent / 'samples' / 'robot-locations.json'
    with robots_file_path.open('r') as f:
        robots = json.load(f)

    for robots_item in robots:
        for i in range(10):
            item = dict(robots_item)
            topic_name = f'robot-locations-{i}'
            item['robot_id'] = str(i)
            item['x'] += i
            item['y'] += 10 - i
            print(item)
            kafka_producer.send(topic_name, value=json.dumps(item).encode('utf8'))

    for i in range(10):
        kafka_producer.send(f'robot-locations-{i}', value=json.dumps(asdict(EndOfStream())).encode('utf8'))
    kafka_producer.flush()


def recreate_topics(topics: List[str]):
    kafka_admin_client = KafkaAdminClient(
        bootstrap_servers='kafka:9092',
    )
    try:
        kafka_admin_client.delete_topics(topics)
    except Exception as exc:
        pass
    sleep(0.1)
    try:
        kafka_admin_client.create_topics([NewTopic(name=topic, num_partitions=1, replication_factor=1) for topic in topics])
    except Exception as exc:
        print(exc)


if __name__ == "__main__":
    window_parallel_streams()



from typing import TypeVar

StreamType = TypeVar('StreamType')
JoinedStreamType = TypeVar('JoinedStreamType')
ResultStreamType = TypeVar('ResultStreamType')


class AliasType:
    def __init__(self, origin_cls, alias):
        self.origin_cls = origin_cls
        self.alias = alias
        self.__name__ = alias

    def __hash__(self):
        return hash(self.alias)

    def __eq__(self, other):
        return self.alias == other.alias

    def __call__(self, *args, **kwargs):
        return self.origin_cls(*args, **kwargs)


def type_alias(stream_type, alias: str):
    return AliasType(stream_type, alias)



import json
import os
from dataclasses import dataclass
from pathlib import Path

from watermill.message_brokers.json_file_message_broker import JsonFileMessageBroker
from watermill.mill import WaterMill


@dataclass(frozen=True, eq=True)
class GoodsItem:
    title: str
    quantity: int
    price: int


@dataclass(frozen=True, eq=True)
class UpdatedGoodsItem:
    title: str
    quantity: int
    price: int


def update_goods_price(goods_item: GoodsItem) -> UpdatedGoodsItem:
    return UpdatedGoodsItem(
        title=goods_item.title,
        quantity=goods_item.quantity,
        price=int(goods_item.price * 1.1)
    )


def single_stream_example_with_result():
    input_file_name = os.path.join(os.path.dirname(__file__), 'samples/goods.json')
    output_file_name = '/tmp/updated-goods.json'

    message_broker = JsonFileMessageBroker({
        GoodsItem: Path(input_file_name),
        UpdatedGoodsItem: Path(output_file_name),
    })

    mill = WaterMill(
        broker=message_broker,
        process_func=update_goods_price,
        stream_cls=GoodsItem,
    )
    mill.run()

    with open(input_file_name, 'r') as input_file:
        input_items = json.load(input_file)

    with open(output_file_name, 'r') as result_file:
        updated_items = json.load(result_file)

    print(f'Input stream:   \n{json.dumps(input_items, sort_keys=True)}')
    print(f'Result stream:  \n{json.dumps(updated_items, sort_keys=True)}')


if __name__ == "__main__":
    single_stream_example_with_result()



import json
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from logging import getLogger
from time import time
from typing import Any
from typing import Callable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Type

from kafka import KafkaConsumer
from kafka import KafkaProducer
from kafka import SimpleClient
from kafka import TopicPartition
from kafka.structs import OffsetCommitRequestPayload

from watermill.message_brokers.message_broker import DataKey
from watermill.message_brokers.message_broker import EndOfStream
from watermill.message_brokers.message_broker import EndOfStreamError
from watermill.message_brokers.message_broker import MessageBroker
from watermill.message_brokers.message_broker import RightJoinedStreamRanAhead
from watermill.message_brokers.message_broker import StreamTimeoutError
from watermill.message_brokers.message_broker import StreamType
from watermill.message_brokers.message_broker import StreamsShift


logger = getLogger(__name__)


@dataclass
class ConsumerTopicOffset:
    consumer_group_name: str
    topic_name: str
    offset: int


@dataclass
class KafkaDataKey(DataKey):
    consumer_offsets: List[ConsumerTopicOffset] = field(default_factory=list)


class KafkaMessageBroker(MessageBroker):
    def __init__(
            self,
            topic_names: Mapping[Type[StreamType], str],
            kafka_url: str,
            data_timeout: Optional[int] = None,
            poll_timeout: Optional[int] = None,
            consumer_group_name: Optional[str] = None,
    ):
        assert topic_names

        self._kafka_url = kafka_url
        self._kafka_client = SimpleClient(self._kafka_url)
        self._serializers = {}
        self._topic_names = topic_names
        self._consumers = {}
        self._producer = self._initialize_producer()
        self._data_timeout_ms = data_timeout or 60 * 60 * 1000
        self._poll_timeout_ms = poll_timeout
        if not self._poll_timeout_ms:
            self._poll_timeout_ms = self._data_timeout_ms
        self._consumer_group_name = consumer_group_name or 'watermill'
        self._partitions = {}

    def get_elements(
            self,
            element_type: Type[StreamType],
            comparator: Callable[[Any, Any], StreamsShift] = None
    ) -> dict:
        if element_type not in self._consumers:
            group_id = element_type.__name__
            if self._consumer_group_name:
                group_id = self._consumer_group_name + '-' + group_id
            self._consumers[element_type] = self._initialize_consumer(
                topic_name=self._topic_names[element_type],
                group_id=group_id,
                stream_type=element_type,
            )

        consumer = self._consumers[element_type]
        element = self._get_next_element(element_type, consumer)
        if not comparator:
            yield element
            return

        while comparator(element) == StreamsShift.Greater:
            logger.warning(f'Skipped stream {element_type} element')
            element = self._get_next_element(element_type, consumer)

        if comparator(element) == StreamsShift.Less:
            self._step_back(element_type)
            raise RightJoinedStreamRanAhead(element_type)

        while comparator(element) == StreamsShift.Equal:
            yield element
            try:
                element = self._get_next_element(element_type, consumer)
            except EndOfStreamError:
                self._step_back(element_type)
                raise

        self._step_back(element_type)

    def commit(self, data_key: Optional[DataKey] = None):
        assert self._consumers
        assert data_key is None or isinstance(data_key, KafkaDataKey)
        if not data_key:
            for stream_type in self._topic_names.keys():
                if stream_type in self._consumers:
                    consumer = self._consumers[stream_type]
                    logger.debug(f'committing {stream_type} stream')
                    consumer.commit()
                    partition = self._partitions[stream_type]
                    logger.debug(f'{stream_type} stream committed offset: {consumer.position(partition)}')
        else:
            for consumer_topic_offset in data_key.consumer_offsets:
                self._kafka_client.send_offset_commit_request(
                    consumer_topic_offset.consumer_group_name,
                    [OffsetCommitRequestPayload(consumer_topic_offset.topic_name, 0, consumer_topic_offset.offset, None)]
                )

    def send(self, item_type: Type[StreamType], item: dict):
        self._producer.send(self._topic_names[item_type], json.dumps(item, ensure_ascii=False).encode('utf8'))
        self._producer.flush()

    def send_eos(self, item_type: Type[StreamType]):
        self._producer.send(self._topic_names[item_type], json.dumps(asdict(EndOfStream()), ensure_ascii=False).encode('utf8'))
        self._producer.flush()

    def get_data_key(self) -> Optional[DataKey]:
        consumer_offsets = []
        for element_type, consumer in self._consumers.items():
            group_id = element_type.__name__
            if self._consumer_group_name:
                group_id = self._consumer_group_name + '-' + group_id

            partition = self._partitions[element_type]
            consumer_offsets.append(ConsumerTopicOffset(
                consumer_group_name=group_id,
                topic_name=self._topic_names[element_type],
                offset=consumer.position(partition)
            ))
        return KafkaDataKey(consumer_offsets=consumer_offsets)

    def _initialize_consumer(self, topic_name: str, group_id: str, stream_type: Type[StreamType]) -> KafkaConsumer:
        consumer =  KafkaConsumer(
            bootstrap_servers=self._kafka_url,
            group_id=group_id,
            auto_offset_reset='earliest',
            enable_auto_commit=False,
            max_poll_records=1,
        )
        partition = TopicPartition(topic_name, 0)
        self._partitions[stream_type] = partition
        consumer.assign([partition])
        logger.debug(f'{topic_name} topic startup offset: {consumer.position(partition)}')
        return consumer

    def _initialize_producer(self) -> KafkaProducer:
        return KafkaProducer(bootstrap_servers=self._kafka_url)

    def _get_next_element(self, element_type: Type[StreamType], consumer: KafkaConsumer) -> dict:
        timeout = self._poll_timeout_ms
        params = {'timeout_ms': timeout} if timeout else {}
        with _TimerContext() as timer:
            next_elements = consumer.poll(**params)
        if not next_elements:
            if timer.elapsed >= self._data_timeout_ms or self._poll_timeout_ms is None:
                raise EndOfStreamError(stream_type=element_type)

            raise StreamTimeoutError(stream_type=element_type)

        for topic, list_of_records in next_elements.items():
            assert len(list_of_records) == 1
            record = list_of_records[0]
            data_bytes = record.value
            element = json.loads(data_bytes.decode('utf8'))

        if element.get('eos__'):
            raise EndOfStreamError(stream_type=element_type)
        return element

    def _step_back(self, element_type: Type[StreamType]):
        consumer = self._consumers[element_type]
        partition = self._partitions[element_type]
        offset = consumer.position(partition)
        consumer.seek(partition, offset - 1)
        logger.debug(f'seek backward 1 message for {element_type}')


class _TimerContext:
    def __init__(self):
        self._start_time = 0
        self._end_time = 0

    def __enter__(self):
        self._start_time = self._end_time = time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._end_time = time()

    @property
    def elapsed(self):
        return self._end_time - self._start_time
